""" Stores most widgets, which are located on ytcon's main screen """
import urwid
from log import logger
from control.control import ControlClass
from settings.settings_processor import settings
from widgets.input_handler import InputHandler
from widgets.top_pile import widgets_tp

from render.static_methods import render_static
from render.colors import colors

class Widgets:
	""" It stores main widgets inside it. """
	def __init__(self):
		self.log_widget = urwid.Text("Initializing, please wait")
		self.error_widget = urwid.Text("Initializing, please wait")
		self.input_widget = InputHandler.InputBox("Enter URL > ")

		self.main_settings_button = urwid.Button("Settings", on_press=settings.show_settings_call)
		self.main_clear_button = urwid.Button("Clear", on_press=self.main_clear_button_ON_PRESS)
		self.main_footer_clipboard_autopaste_button = urwid.Button("Autopaste", on_press=settings.clipboard_autopaste_switch)

		self.main_footer_buttons = urwid.GridFlow(
			[self.main_settings_button, self.main_clear_button, self.main_footer_clipboard_autopaste_button],
			cell_width=13, h_sep=2, v_sep=1, align="left")
		logger.debug(self.main_footer_buttons.contents)
		self.main_footer_buttons_with_attrmap = urwid.AttrMap(self.main_footer_buttons, "buttons_footer")

		self.auto_update_avalible_text_indicator = urwid.Text("- - -")

		self.main_footer = urwid.Pile(
				[
				self.error_widget,
				urwid.Text("- - -"),
				self.log_widget,
				urwid.Text("- - -"),
				self.input_widget,
				urwid.Divider(),
				self.auto_update_avalible_text_indicator,
				self.main_footer_buttons_with_attrmap,
				])
		self.main_widget = urwid.Frame(
			urwid.Filler(widgets_tp.top_pile, "top"),
			footer=self.main_footer,
			focus_part='footer')

	def main_clear_button_ON_PRESS(self, _=None): # TODO: ?
		""" Funcion that called by < Clear > button press """
		ControlClass.clear()
		render_static.flash_button_text(self.main_clear_button, colors.light_yellow, 2)

widgets = Widgets()



"""
	The module which contains the settings category classes.

	They will be automatically found and placed in a special dynamic list (self.settings_sections)
	that will be shown to the user
"""

# TODO: Please consider rewrite to dynamic plug-ins system
import urwid

from control.variables import variables
from render.colors import colors

from render.render import render
RenderClass = render

from control.control import ControlClass

from settings.settings_processor import settings

class SettingsSections:
	"""
	The module which contains the settings category classes.

	They will be automatically found and placed in a special dynamic list (self.settings_sections)
	that will be shown to the user
	"""
	def __init__(self):
		# Get all class attributes (sections)
		class_attributes = vars(SettingsSections)

		# Filter only classes
		self.settings_sections = [cls for cls in class_attributes.values() if isinstance(cls, type)]

		#print(self.settings_sections)

	class General_SECTION: # pylint: disable=attribute-defined-outside-init # because get() initializes a class
		""" General settings section """
		name = "General"

		def get(self):
			""" Get content of section """
			self.settings_checkbox_clipboard = urwid.CheckBox("Clipboard auto-paste", on_state_change=settings.setting_switch, user_data="clipboard_autopaste")

			# UPDATE CHECKBOXES
			self.update()

			settings_pile = urwid.Pile([
				urwid.Divider(),
				self.settings_checkbox_clipboard,
				urwid.Divider(),
				])
			return settings_pile

		def update(self):
			""" Update checkbox states for they don't lie """
			self.settings_checkbox_clipboard.set_state(settings.get_setting("clipboard_autopaste"), do_callback=False)

	class Appearance_SECTION: # pylint: disable=attribute-defined-outside-init # because get() initializes a class
		""" settings section related to appearance """
		name = "Appearance"
		def get(self):
			""" Get content of section """
			self.settings_checkbox_progresstype_detailed = urwid.CheckBox([
				(colors.cyan, "46% |███▍   | - Detailed"),
				"\nUse some unicode characters (▏;▍;▋;▉;█)\nto display the percentage more accurately.\nDoesn't fully work in tty",
				], on_state_change=settings.setting_change_content, user_data=("progressbar_appearance", "detailed"))
			self.settings_checkbox_progresstype_simple = urwid.CheckBox([
				(colors.cyan, "46% |████   | - Simple"),
				"\nUse only ACSII squares (█) to show percentage"
				], on_state_change=settings.setting_change_content, user_data=("progressbar_appearance", "simple"))
			self.settings_checkbox_progresstype_arrow = urwid.CheckBox([
				(colors.cyan, "46% |===>   | - Arrow"),
				"\nLet's just add some oldfag style 😎"
				], on_state_change=settings.setting_change_content, user_data=("progressbar_appearance", "arrow"))
			self.settings_checkbox_progresstype_pacman = urwid.CheckBox([
				(colors.cyan, "46% |--C o | - Pacman"),
				"\nPacman game"
				], on_state_change=settings.setting_change_content, user_data=("progressbar_appearance", "pacman"))

			# UPDATE CHECKBOXES
			self.update()

			settings_pile = urwid.Pile([
				urwid.Divider(),
				urwid.Text((colors.light_yellow, "Progress bar type")),
				self.settings_checkbox_progresstype_detailed,
				urwid.Divider(),
				self.settings_checkbox_progresstype_simple,
				urwid.Divider(),
				self.settings_checkbox_progresstype_arrow,
				urwid.Divider(),
				self.settings_checkbox_progresstype_pacman,
				urwid.Divider(),
				])

			return settings_pile

		def update(self):
			""" Update checkbox states for they don't lie """
			self.settings_checkbox_progresstype_detailed.set_state(settings.get_setting("progressbar_appearance") == "detailed", do_callback=False)
			self.settings_checkbox_progresstype_simple.set_state(settings.get_setting("progressbar_appearance") == "simple", do_callback=False)
			self.settings_checkbox_progresstype_arrow.set_state(settings.get_setting("progressbar_appearance") == "arrow", do_callback=False)
			self.settings_checkbox_progresstype_pacman.set_state(settings.get_setting("progressbar_appearance") == "pacman", do_callback=False)


	class Fetching_SECTION: # pylint: disable=attribute-defined-outside-init # because get() initializes a class
		""" Fetching settings section - related to yt-dlp downloding """
		name = "Fetching"
		def get(self):
			""" Get content of section """
			self.settings_checkbox_sp = urwid.CheckBox([(colors.light_yellow, "\"Special mode\""), "\nUse different user-agent and extract cookies from chromium"], on_state_change=settings.setting_switch, user_data="special_mode")
			self.settings_checkbox_nocert = urwid.CheckBox([(colors.light_yellow, "Do not check website certificates"), "\nEnable this if \"SSL: CERTIFICATE_VERIFY_FAILED\" error occurs"], on_state_change=settings.setting_switch, user_data="no_check_certificate")

			# UPDATE CHECKBOXES
			self.update()

			settings_pile = urwid.Pile([
				urwid.Divider(),
				self.settings_checkbox_sp,
				urwid.Divider(),
				self.settings_checkbox_nocert,
				urwid.Divider(),
				])

			return settings_pile

		def update(self):
			""" Update checkbox states for they don't lie """
			self.settings_checkbox_sp.set_state(settings.get_setting("special_mode"), do_callback=False)
			self.settings_checkbox_nocert.set_state(settings.get_setting("no_check_certificate"), do_callback=False)

	class Playlists_SECTION: # pylint: disable=attribute-defined-outside-init # because get() initializes a class
		""" Playlist settings section """
		name = "Playlists"
		def get(self):
			""" Get content of section """
			self.settings_checkbox_ignerr = urwid.CheckBox([
				(colors.light_yellow, "Ignore downloading errors"),
				(colors.light_red, "\n<!!> Dangerous option - makes ytcon a little unstable\nPlease use only if necessary <!!>"),
				"\nUse this so as not to interrupt the download if\none of the video in the playlist is not available"
				], on_state_change=settings.setting_switch, user_data="ignoreerrors")

			# UPDATE CHECKBOXES
			self.update()

			settings_pile = urwid.Pile([
				urwid.Divider(),
				self.settings_checkbox_ignerr,
				urwid.Divider(),
				])

			return settings_pile

		def update(self):
			""" Update checkbox states for they don't lie """
			self.settings_checkbox_ignerr.set_state(settings.get_setting("ignoreerrors"), do_callback=False)

	class Debug_SECTION: # pylint: disable=attribute-defined-outside-init # because get() initializes a class
		""" DEBUG settings section """
		name = "Debug"
		def get(self):
			""" Get content of section """
			self.settings_checkbox_delete_af = urwid.CheckBox("Delete after download", on_state_change=ControlClass.delete_after_download_switch)

			# UPDATE CHECKBOXES
			self.update()

			settings_pile = urwid.Pile([
				urwid.Divider(),
				urwid.Text((colors.light_red, "The settings found here are made for testing purposes!")),
				urwid.Text((colors.light_red, "Changing these settings is not recommended.")),
				urwid.Divider(),
				urwid.Text((colors.light_red, "Also, Debug settings WILL NOT be saved when you click on the \"Save to config file\" button")),
				urwid.Divider(),
				urwid.Text("- = -"),
				urwid.Divider(),
				self.settings_checkbox_delete_af,
				urwid.Divider(),
				])

			return settings_pile

		def update(self):
			""" Update checkbox states for they don't lie """
			self.settings_checkbox_delete_af.set_state(variables.delete_after_download, do_callback=False)

	# = - E X A M P L E - =
	#class Three_SECTION:
	#	# Test section
	#	name = "3"
	#	def get(self):
	#		# Get content of section
	#		return urwid.Text('helo3')

settings_sections = SettingsSections()



""" The module that is responsible for the operation of the clipboard in ytcon """
import re
import sys
import time
import threading

import clipman

from log import journal, logger

from control.variables import variables
from control.exit import exit_with_exception, traceback

from settings.settings_processor import settings
from settings_menu.render import update_checkboxes

from downloader.main import downloader
#from render.loop import loop_container

# TODO: INIT NEEDS REWRITING.
# - also, separate control class for clipboard,
#   not variables.clipboard_checker_state_launched

# 	↓ shit
def clipboard_init():
	""" Initializes clipman in ytcon boot stage. If ytcon already booted, it clipman inits in clipboard_checker() """
	try:
		clipman.init()
	except Exception as e: # pylint: disable=broad-except
		logger.info(traceback.format_exc())
		print("[!!] An clipboard error occurred!\n")
		print(f"- {type(e).__name__}: {e}")
		print("\nYou can follow instructions in this error message, or ignore it")
		print("BUT, if you ignore it, clipboard auto-paste will be unavalible.\n")
		print("Also, if this error message doesn't contain instructions,")
		print("and does not contain any understandable text for your human language, please make an issue")
		print("https://github.com/NikitaBeloglazov/clipman/issues/new")
		print("Full traceback can be found in info.log\n")

		try:
			user_answer = input("Ignore it? [y/N] > ")
		except KeyboardInterrupt:
			print("Exiting..")
			sys.exit(1)

		if user_answer.lower() in ("yes", "y"):
			journal.error("[YTCON] If you don't want answer \"yes\" every time, solve the problem, or disable auto-paste in settings and PRESS \"Save to config file\"")
			settings.write_setting("clipboard_autopaste", False)
		else:
			print("Exiting..")
			sys.exit(1)

url_regex = r"^(https?:\/\/)?([\w-]{1,32}\.[\w-]{1,32})[^\s@]*$"
def clipboard_checker():
	"""
	Checks the clipboard for new entries against old ones.
	If it sees new content on the clipboard, it will check whether this is a site, if it detects site, download starts
	"""

	# Set the button yellow and DO NOT start daemon
	variables.clipboard_checker_state_launched = "Do not start"

	if clipman.dataclass.init_called is False:
		try:
			clipman.init()
		except:
			logger.info(traceback.format_exc())
			journal.error("[YTCON] An error occurred while initializing the clipboard. You can see the error in info.log. Or save the Auto-paste option enabled in the config file, restart ytcon, and after that you will see an error with detailed instructions.")

			# Keep setting ON for "save to config file" ability
			time.sleep(60)
			settings.write_setting("clipboard_autopaste", False)
			update_checkboxes()
			variables.clipboard_checker_state_launched = False
			return None

	try:
		variables.clipboard_checker_state_launched = True
		journal.info("[YTCON] Clipboard auto-paste is ON.")

		old_clip = ""

		while True:
			if settings.get_setting("clipboard_autopaste") is False:
				variables.clipboard_checker_state_launched = False
				journal.info("[YTCON] Clipboard auto-paste turned off.")
				return None

			new_clip = clipman.paste()
			if new_clip != old_clip:
				if re.fullmatch(url_regex, new_clip):
					journal.info("[CLIP] New URL detected: " + new_clip)
					threading.Thread(target=downloader, args=(new_clip,), daemon=True).start()
				else:
					logger.debug("clipboard content: %s", new_clip)
					journal.info("[CLIP] New clipboard content detected. But this is not URL. Ignoring..")
			old_clip = new_clip
			time.sleep(1)
	except:
		exit_with_exception(str(traceback.format_exc()))
		return None



""" The module that is responsible for rendering some things the settings menu """
import traceback
import urwid

from log import logger
from render.loop import loop_container
from render.render import render
RenderClass = render

from widgets.main_widgets import widgets

from control.exit import exit_with_exception, traceback

from settings.settings_processor import settings
from settings_menu.variables import settings_menu_variables
from settings_menu.sections import settings_sections

from app_update import app_updates


def update_checkboxes():
	"""
	!LEGACY!: update the checkboxes so that their status is not a lie
	"""
	if settings_menu_variables.settings_show is True:
		sett.update()

def gen_SimpleFocusListWalker_with_footer(contents, footer, width=20):
	"""
	Some analogue of urwid.Frame, it contains a body (contents) and footer,
	but at the same time we CAN switch the focus between them
	"""
	# Count body (contents) rows
	contents_rows = 0
	for i in contents:
		contents_rows = contents_rows + i.rows((width,))

	# Count footer rows
	footer_rows = 0
	for i in footer:
		footer_rows = footer_rows + i.rows((width,))

	filler_height = RenderClass.height - contents_rows - footer_rows
	filler_list = []

	# Filling the empty space between widgets using a urwid.Divider
	for i in range(0, filler_height):
		filler_list.append(urwid.Divider())
	return urwid.Pile(contents + filler_list + footer)

class SettingsRenderClass:
	""" The class that is responsible for rendering the settings menu """
	def __init__(self):
		self.exit_settings_button = urwid.Button("Exit from settings", on_press=settings.show_settings_call)
		self.save_settings_button = urwid.Button("Save to config file", on_press=settings.save)
		self.load_settings_button = urwid.Button("Load from config file", on_press=settings.load)

		self.footer_widget = urwid.Pile([
			widgets.error_widget,
			urwid.Text("- - -"),
			widgets.log_widget,
		])

		# - =
		# just placeholders. nevermind
		self.columns = None
		self.right_widget = None
		# - =

		# - = - Section buttons mapping - = - = - = - = - = - = -
		self.connected_sections = settings_sections.settings_sections

		self.section_buttons = [
				urwid.AttrMap(urwid.Text(" - = Categories = -"), "green_background", ""),
				]

		for i in self.connected_sections:
			self.section_buttons.append(
				urwid.AttrMap(
					urwid.Button(i.name, on_press=self.set_right_section, user_data=i),
					"", "reversed"
				)
			)
		# - = - = - = - = - = - = - = - = - = - = - = - = - = - =

		self.left_widget_sflw = gen_SimpleFocusListWalker_with_footer(
			self.section_buttons,
			[
				urwid.AttrMap(self.load_settings_button, "cyan", "reversed"),
				urwid.AttrMap(self.save_settings_button, "cyan", "reversed"),
				urwid.Divider(),
				urwid.AttrMap(self.exit_settings_button, "light_cyan", "cyan_background"),
			]
			)

		self.left_widget = urwid.Filler(self.left_widget_sflw, valign="top")

		self.vertical_divider = urwid.Filler(urwid.Text(" " * 100))
		self.set_right_section(None, self.connected_sections[0], update=False)

	def set_right_section(self, _, section, update=True):
		""" A function that puts the specified section class on the right visible part of the interface """
		self.current_section = section
		if update:
			self.update()

	def soft_update(self):
		""" Update current section flags states without re-rendering it """
		try:
			self.current_section_initialized.update()
		except AttributeError:
			logger.debug("soft_update unsucceful because SettingsRenderClass doesn't have initialized settings section")

	def update(self):
		""" re-generate + re-render right visible part of the interface """
		self.current_section_initialized = self.current_section()
		self.right_widget = urwid.Frame(
			urwid.Padding(urwid.Filler(self.current_section_initialized.get(), valign='top'), left=2, right=2, align='center'),

			footer = urwid.Pile([
				app_updates.settings_version_text,
				urwid.LineBox(
					self.footer_widget,
					tlcorner='╭', trcorner='╮', # Rounding corners
					blcorner='', bline='', brcorner='' # Remove bottom line
					),
				]),

			header = urwid.AttrMap(urwid.Text(" - = " + self.current_section.name), "reversed", "") )

		# Create Columns for split screen
		self.columns = urwid.Columns(
			[
			# ALL INCOMING WIDGETS MUST BE BOX
			("fixed", 20, self.left_widget),
			("fixed", 1, self.vertical_divider),
			#("fixed", 1, urwid.AttrMap(self.vertical_divider, "reversed")),
			self.right_widget
			])

		loop_container.loop.widget = self.columns

	def tick_handler_settings(self, _, _1):
		""" Same as tick_handler, but responsible only for settings menu """
		if settings_menu_variables.settings_show is True:
			lol = sett.left_widget_sflw.focus_position - 1 # -1 because header is widget too
			if not lol > len(self.connected_sections)-1: # prevent crash on bottom buttons selection, -1 because len makes +1
				if self.current_section != self.connected_sections[lol]:
					self.set_right_section(None, self.connected_sections[lol])

		# - = - = - = - = - = - = - = - = -
		# Settings page show handler
		if settings_menu_variables.settings_show is True and settings_menu_variables.settings_showed is False:
			try:
				# - = - = -
				# Return to default position
				self.left_widget_sflw.set_focus(1)
				self.set_right_section(None, self.connected_sections[0], update=False)
				# - = - = -
				self.update()
				settings_menu_variables.settings_showed = True
			except:
				exit_with_exception(traceback.format_exc())
		if settings_menu_variables.settings_show is False and settings_menu_variables.settings_showed is True:
			try:
				loop_container.loop.widget = widgets.main_widget
				settings_menu_variables.settings_showed = False
			except:
				exit_with_exception(traceback.format_exc())
		# - = - = - = - = - = - = - = - = -

		# - = - = - = - = - = - = - = - = -
		# Soft checkbox updater
		if settings_menu_variables.settings_soft_update_scheduled is True:
			self.soft_update()
			settings_menu_variables.settings_soft_update_scheduled = False
		# - = - = - = - = - = - = - = - = -

		loop_container.loop.set_alarm_in(0.1, self.tick_handler_settings)

sett = SettingsRenderClass()



""" This module contains methods for drawing progress bars """
import time
from tqdm import tqdm

from settings.settings_processor import settings

class ProgressBarDefs:
	""" Minor methods mostly needed by render_tasks """
	def name_shortener(self, name, symbols):
		""" Shortens filenames so they fit in the console """
		if len(name) < symbols:
			return name
		return name[0:symbols-3].strip() + "..."

	def bettersize(self, text):
		""" Rounds up file sizes """
		if text == "NaN":
			return "NaN"
		if len(text.split(".")) == 1:
			return text
		return text.split(".")[0] + text[-3:-1] + text[-1]

	def divide_without_remainder(self, num):
		"""
		Division without remainder. Used for centering in the whitespace_stabilization function
			print(divide(22))  # Out: [11, 11]
			print(divide(23))  # Out: [11, 12]
		"""
		quotient = num // 2
		remainder = num % 2
		return [quotient, quotient + remainder]

	def whitespace_stabilization(self, text, needed_space):
		""" Reserves space for a certain number of spaces, and centers the information within a line,
		using divide_without_remainder to count the free space.

		If the free space cannot be divided entirely, then the text is centered slightly to the left
		"""
		if len(text) == needed_space:
			return text
		if len(text) > needed_space:
			return text[0:needed_space-2] + ".."
		white_space = needed_space - len(text)
		white_space = self.divide_without_remainder(white_space)

		return ' '*white_space[0] + text + ' '*white_space[1]

	def progressbar_generator(self, percent, error=False):
		""" Generates progress bar """
		percent = int(percent.split(".")[0])
		progress = round(percent / 4)
		white_space = 25 - progress
		if error:
			return f"|{'#'*progress}{' '*white_space}|"

		style = settings.get_setting("progressbar_appearance")

		if style == "detailed":
			return tqdm.format_meter(percent, 100, 0, ascii=False, bar_format="|{bar}|", ncols=27, mininterval=0, maxinterval=0)

		if style == "simple":
			return f"|{'█'*progress}{' '*white_space}|"

		if style == "arrow":
			temp1 = '='*progress
			if temp1 != "":
				temp1 = temp1[:-1] + ">" # replace last symbol to >
			return f"|{temp1}{' '*white_space}|"

		if style == "pacman":
			# [---------Co o o o o o o ]
			#            ^ background ^

			mseconds = time.time() - round(time.time())
			if mseconds > 0: # pylint: disable=simplifiable-if-statement # no, because bool(-1) is True
				mystery_bool = True
			else:
				mystery_bool = False

			progressline = '-'*progress
			if progress not in (0, 25):
				if mystery_bool:
					progressline = progressline[:-1] + "C" # replace last symbol to c
				else:
					progressline = progressline[:-1] + "c" # replace last symbol to C

			# Skip background generation if this not needed
			if progress == 25:
				return f"[{progressline}]"

			# - = Background generation = -
			if progress % 2:
				background = "o"
			else:
				background = " "

			while len(background) != white_space:
				if background[-1] == " ":
					background = background + "o"
				else:
					background = background + " "
			# - = - = - = - = - = - = - = -
			return f"[{progressline}{background[:white_space]}]"
		return None

progressbar_defs = ProgressBarDefs()



""" Сontains settings data and methods for them """
import os
import sys
import pickle
import traceback
from pathlib import Path

from log import journal, logger
from control.variables import variables
from render.colors import colors
from render.static_methods import render_static
from settings_menu.variables import settings_menu_variables

class SettingsClass:
	""" Сontains settings data and methods for them """
	def __init__(self, configpath):

		# Default settings
		self.settings = {
			"special_mode": False,
			"clipboard_autopaste": True,
			"no_check_certificate": False,
			"ignoreerrors": False,
			"progressbar_appearance": "detailed",
			}
		self.configpath = configpath

	class SettingNotFoundError(Exception):
		""" Called if the specified setting is not found (see def get_setting) """

	def show_settings_call(self, _=None):
		""" Settings display state switch, made for urwid.Button(on_press=show_settings_call) """
		settings_menu_variables.settings_show = not settings_menu_variables.settings_show

	def get_setting(self, setting_name):
		""" Get setting, if it not found, calls SettingNotFoundError """
		try:
			return self.settings[setting_name]
		except KeyError as exc:
			raise self.SettingNotFoundError from exc

	def write_setting(self, setting_name, setting_content):
		""" Writes the settings to the memory. Made for the possible use of some "hooks" in the future """
		self.settings[setting_name] = setting_content

	def save(self, button=None): # in the second argument urwid puts the button of which the function was called
		""" Uses pickle for saving settings from memory to ~/.config/settings.db"""
		logger.debug(Path(self.configpath).mkdir(parents=True, exist_ok=True)) # Create dirs if they don't already exist
		with open(self.configpath + "settings.db", "wb") as filee:
			pickle.dump(self.settings, filee)
		journal.info(f"[YTCON] {self.configpath}settings.db saved")
		render_static.flash_button_text(button, colors.green)

	def load(self, button=None): # in the second argument urwid puts the button of which the function was called
		""" Uses pickle for loading settings from ~/.config/settings.db to memory """
		try:
			with open(self.configpath + "settings.db", "rb") as filee:
				self.settings.update(pickle.load(filee))
			journal.info(f"[YTCON] {self.configpath}settings.db loaded")
			settings_menu_variables.settings_soft_update_scheduled = True # update checkboxes
			self.update_ydl_opts()
			render_static.flash_button_text(button, colors.green)
		except FileNotFoundError:
			# If file not found
			journal.warning(f"[YTCON] Saved settings load failed: FileNotFoundError: {self.configpath}settings.db")
			render_static.flash_button_text(button, colors.red)
		except EOFError as exc:
			# If save file is corrupted
			logger.debug(traceback.format_exc())
			journal.warning(f"[YTCON] Saved settings load FAILED: EOFError: {exc}: {self.configpath}settings.db")
			journal.error("[YTCON] YOUR SETTINGS FILE IS CORRUPTED. Default settings restored and corrupted save file removed.")
			self.write_setting("clipboard_autopaste", False)
			settings_menu_variables.settings_soft_update_scheduled = True # update checkboxes
			journal.warning("[YTCON] Clipboard autopaste has been turned off for security reasons. You can it enable it in settings")
			logger.debug(os.remove(f"{self.configpath}settings.db"))

	def clipboard_autopaste_switch(self, _=None, _1=None):
		""" Clipboard autopaste switch function for urwid.Button's. FOR BACK COMPABILITY """
		self.setting_switch(None, None, name="clipboard_autopaste")

	def setting_switch(self, _=None, state=None, name=None):
		""" Switches state to negative current state or to state set by state argument. Made for for urwid.Button's """
		if name is None:
			raise TypeError

		if state is None:
			state = not self.get_setting(name)

		journal.info("")
		journal.info(f"[YTCON] {name}: {self.get_setting(name)} -> {state}")
		self.write_setting(name, state)
		self.update_ydl_opts()
		settings_menu_variables.settings_soft_update_scheduled = True

	def setting_change_content(self, _=None, _1=None, data=None):
		"""
		Change content in setting where negative state cannot be determined.
		Made for for urwid.Button's with pre-writed arguments
		"""
		if data is None:
			raise TypeError

		name = data[0]
		set_data = data[1]

		journal.info("")
		journal.info(f"[YTCON] {name}: {self.get_setting(name)} -> {set_data}")
		self.write_setting(name, set_data)
		self.update_ydl_opts()
		settings_menu_variables.settings_soft_update_scheduled = True

	def update_ydl_opts(self):
		""" Updates some setting-related ydl_opts. Maybe something like post-change scripts? """
		#journal.info(pprint.pformat(variables.ydl_opts))
		#journal.info("updated ydl_opts")

		# - = Special mode cookie extractor activator = -
		if settings.get_setting("special_mode") is True and "cookiesfrombrowser" not in variables.ydl_opts:
			variables.ydl_opts["cookiesfrombrowser"] = ('chromium', ) # needed for some sites with login only access. you may need to replace it with the correct one
		elif settings.get_setting("special_mode") is False and "cookiesfrombrowser" in variables.ydl_opts:
			del variables.ydl_opts["cookiesfrombrowser"]
		# - = - = - = - = - = - = - = - = - = - = - = - =

		# - = Certificates ignore activator = -
		if settings.get_setting("no_check_certificate") is True and "nocheckcertificate" not in variables.ydl_opts:
			variables.ydl_opts["nocheckcertificate"] = True
		elif settings.get_setting("no_check_certificate") is False and "nocheckcertificate" in variables.ydl_opts:
			del variables.ydl_opts["nocheckcertificate"]
		# - = - = - = - = - = - = - = - = - = - = - = - =

		# - = Certificates ignore activator = -
		if settings.get_setting("ignoreerrors") is True and "ignoreerrors" not in variables.ydl_opts:
			variables.ydl_opts["ignoreerrors"] = True
		elif settings.get_setting("ignoreerrors") is False and "ignoreerrors" in variables.ydl_opts:
			del variables.ydl_opts["ignoreerrors"]
		# - = - = - = - = - = - = - = - = - = - = - = - =

		#journal.info(pprint.pformat(variables.ydl_opts))

# - - - - - - - - - - - - -
# Save file folder check
if "XDG_CONFIG_HOME" in os.environ:
	configpath = os.path.expanduser(os.environ["XDG_CONFIG_HOME"] + "/ytcon/")
else:
	configpath = os.path.expanduser("~/.config/ytcon/")

try:
	Path(configpath).mkdir(parents=True, exist_ok=True)
	with open(configpath + "write_test", "wb") as filee:
		pass
	os.remove(configpath + "write_test")
except:
	print(traceback.format_exc())
	print("= = =\n[!!] An error was occurred!\n")
	print("Save file folder check failed. Maybe XDG_CONFIG_HOME env or dir permissions broken?")
	print("The following path has problems: " + configpath)
	sys.exit(1)

logger.debug("config path: %s", configpath)
# - = - = - = - = - = - = - = - = - = - = - = - = - = - =
settings = SettingsClass(configpath)



""" Tags version in files for publishing in pypi. TODO: Deprecated """
import subprocess
import os

os.system("ls -a")
tag = subprocess.check_output("git describe --tags", shell=True, encoding="UTF-8")

print("[TAG MARKER] git response: " + tag)
tag = tag.replace("\n", "").replace("v", "")
if tag.find("-") > 1:
	tag = tag[0:tag.find("-")]
print("[TAG MARKER] FOUND TAG: " + tag)

print("[TAG MARKER] MARKING VERSION IN pyproject.toml")
# Open the file for reading and writing (the 'r+' mode provides both read and write access)
with open('pyproject.toml', 'r+', encoding="UTF-8") as file:
	# Reading the contents of a file
	content = file.read()
	# Replace the desired part of the text using the replace method
	new_content = content.replace('!!{PLACEHOLDER}!!', tag)
	# Move the file pointer to the beginning (0) to overwrite
	file.seek(0)
	# Write the changed content back to the file
	file.write(new_content)
	# Trim the file if the new content is shorter than the old one to remove the remaining data
	file.truncate()
print("[TAG MARKER] DONE")

print("[TAG MARKER] MARKING VERSION IN src/ytcon/__version__.py")
with open('src/ytcon/__version__.py', 'r+', encoding="UTF-8") as file:
	# Reading the contents of a file
	content = file.read()
	# Replace the desired part of the text using the replace method
	new_content = content.replace('!!{PLACEHOLDER}!!', tag)
	# Move the file pointer to the beginning (0) to overwrite
	file.seek(0)
	# Write the changed content back to the file
	file.write(new_content)
	# Trim the file if the new content is shorter than the old one to remove the remaining data
	file.truncate()
print("[TAG MARKER] DONE")

print("[TAG MARKER] Work finished")



"""
Created to simplify the distribution of parameters,
work is organized here with playlists and requesting several formats on youtube
"""

from control.variables import variables

class MapVariablesClass:
	""" Created to simplify the distribution of parameters, work is organized here with playlists and requesting several formats on youtube """

	def main(self, multiple_formats, infolist, filename):
		""" Finding some specific parameters and using a loop assign if there are several files """
		if multiple_formats:
			variables.queue_list[infolist["original_url"]] = {}
			variables.queue_list[infolist["original_url"]]["meta_index"] = True
			variables.queue_list[infolist["original_url"]]["multiple_formats"] = True
			variables.queue_list[infolist["original_url"]]["formats"] = []
			variables.queue_list[infolist["original_url"]]["status"] = "waiting"
			for i in infolist["requested_formats"]:
				temp1_index = infolist["original_url"] + ":" + i["format_id"]
				variables.queue_list[infolist["original_url"]]["formats"].append(i["format_id"])
				self.map_variables(temp1_index, infolist, i, filename)
			return temp1_index
		# else:
		temp1_index = infolist["original_url"]
		self.map_variables(temp1_index, infolist, infolist, filename)
		return temp1_index

	def map_variables(self, temp1_index, infolist, i, filename):
		""" Main parameter assigner. In some cases, it can be used in a loop """
		variables.queue_list[temp1_index] = {}
		variables.queue_list[temp1_index]["status"] = "waiting"
		variables.queue_list[temp1_index]["status_short_display"] = "Wait"
		variables.queue_list[temp1_index]["percent"] = "0.0%"
		variables.queue_list[temp1_index]["speed"] = "0KiB/s"
		try:
			variables.queue_list[temp1_index]["size"] = str(round(i["filesize"]/1e+6)) + "MiB"
		except KeyError:
			variables.queue_list[temp1_index]["size"] = "???MiB"
		variables.queue_list[temp1_index]["downloaded"] = "0MiB"
		variables.queue_list[temp1_index]["eta"] = "ETA ??:??"
		variables.queue_list[temp1_index]["name"] = infolist["fulltitle"]
		if i["resolution"] == "audio only":
			variables.queue_list[temp1_index]["resolution"] = "audio"
		else:
			if i.get("width", None) is None and i.get("height", None) is None:
				variables.queue_list[temp1_index]["resolution"] = "???х???"
			else:
				variables.queue_list[temp1_index]["resolution"] = (str(i.get("width", None)) + "x" + str(i.get("height", None))).replace("None", "???")
		variables.queue_list[temp1_index]["site"] = infolist["extractor"].lower()
		variables.queue_list[temp1_index]["file"] = filename

	def mark_as_error(self, url):
		""" Change the status of the downloaded link to Error if such link exists """
		if url in variables.queue_list:
			variables.queue_list[url]["status"] = "error"
			if "multiple_formats" in variables.queue_list[url]:
				for i in url["formats"]:
					temp1_index = url + ":" + i
					variables.queue_list[temp1_index]["status"] = "error"
					variables.queue_list[temp1_index]["status_short_display"] = "Error"
			else:
				variables.queue_list[url]["status_short_display"] = "Error"

map_variables = MapVariablesClass()



"""
	tick_handler just checks some conditions every few seconds and executes them.
	Directly not responsible for rendering, but changes some buttons color
"""

import sys
import time
import threading
import traceback

import urwid

from log import journal

from control.variables import variables

from render.colors import colors

from widgets.main_widgets import widgets
from settings.settings_processor import settings

from app_update import app_updates
from misc.clipboard import clipboard_checker

def tick_handler(loop, _):
	""" It just checks some conditions every few seconds and executes them. Directly not responsible for rendering, but changes some buttons color """

	# - = - = - = - = - = - = - = - = -
	# Autopaste button color changer
	if (settings.get_setting("clipboard_autopaste") is True and variables.clipboard_checker_state_launched is not True) or (settings.get_setting("clipboard_autopaste") is False and variables.clipboard_checker_state_launched is not False):
		widgets.main_footer_buttons.contents[2] = (urwid.AttrMap(widgets.main_footer_clipboard_autopaste_button, "yellow"), widgets.main_footer_buttons.contents[2][1])
		variables.temp["autopaste_button_color"] = "yellow" # some kind of cache

	elif variables.clipboard_checker_state_launched is not True and variables.temp["autopaste_button_color"] != "light_red":
		widgets.main_footer_buttons.contents[2] = (urwid.AttrMap(widgets.main_footer_clipboard_autopaste_button, "light_red"), widgets.main_footer_buttons.contents[2][1])
		variables.temp["autopaste_button_color"] = "light_red" # some kind of cache

	elif variables.clipboard_checker_state_launched is True and variables.temp["autopaste_button_color"] != "buttons_footer":
		widgets.main_footer_buttons.contents[2] = (urwid.AttrMap(widgets.main_footer_clipboard_autopaste_button, "buttons_footer"), widgets.main_footer_buttons.contents[2][1])
		variables.temp["autopaste_button_color"] = "buttons_footer" # some kind of cache
	# - = - = - = - = - = - = - = - = -

	# - = Clipboard thread activator = -
	if settings.get_setting("clipboard_autopaste") and variables.clipboard_checker_state_launched is False:
		threading.Thread(target=clipboard_checker, daemon=True).start()
	# - = - = - = - = - = - = - = - = -

	# - = - = - = - = - = - = - = - = -
	# The error handler, if it sees variables.exit = True,
	# then exits the program commenting this with the text from variables.exception.
	# The parent function of such actions: exit_with_exception()
	if variables.exit is True:
		loop.stop()
		print("An unknown error has occurred!\n")
		time.sleep(0.5)
		print(variables.exception)
		sys.exit(1)

	if variables.auto_update_safe_gui_stop is True:
		try:
			loop.stop()
		except:
			journal.debug(traceback.format_exc())

		try:
			app_updates.update_thread.join()
		except KeyboardInterrupt:
			print(" - Okay, canceled")
		sys.exit()
	# - = - = - = - = - = - = - = - = -

	# Prevent focus from remaining on footer buttons after pressing them
	widgets.main_footer.set_focus(widgets.input_widget)

	# - =
	loop.set_alarm_in(0.3, tick_handler)



def tick_handler_big_delay(loop, _):
	""" Same as tick_handler, but with bigger delay. Made for optimization purposes. """

	# - = - = - = - = - = - = - = - = -
	# Draw version in settings
	app_updates.update_settings_version_text()

	# New-update-avalible notificator
	if app_updates.auto_update_avalible is True:
		widgets.auto_update_avalible_text_indicator.set_text((colors.cyan, f"- - -\nAuto update {app_updates.version} -> {app_updates.pypi_version} is avalible! Write \"update\" to easy update right now!"))
	# - = - = - = - = - = - = - = - = -

	# - =
	loop.set_alarm_in(4, tick_handler_big_delay)



""" Contains top_pile with widgets, which used by loops.render_tasks, for indicating current downloads """
import urwid

class Widgets_TopPile_Basic:
	""" Contains top_pile with widgets, which used by loops.render_tasks, for indicating current downloads """
	def __init__(self):
		self.top_pile = urwid.Pile([])

widgets_tp = Widgets_TopPile_Basic()



"""
	A hook that is called every time by yt-dlp when the state of the task changes (example percent changed),
	and the hook writes the necessary information to the class in order to draw it later
"""
import pprint

from log import journal, logger

from control.variables import variables
from control.exit import exit_with_exception, traceback

from misc.ffmpeg import get_resolution_ffprobe

def hook(d):
	""" A hook that is called every time by yt-dlp when the state of the task changes (example percent changed),
	and the hook writes the necessary information to the class in order to draw it later """
	try:
		# - = - = - log spam filter - = - = - = - =
		if "automatic_captions" in d["info_dict"]:
			del d["info_dict"]["automatic_captions"]
		if "formats" in d["info_dict"]:
			del d["info_dict"]["formats"]
		if "thumbnails" in d["info_dict"]:
			del d["info_dict"]["thumbnails"]
		if "heatmap" in d["info_dict"]:
			del d["info_dict"]["heatmap"]

		# ↓ ???
		d["info_dict"]["formats"] = []
		d["info_dict"]["thumbnails"] = []
		d["info_dict"]["subtitles"] = []
		d["info_dict"]["fragments"] = []
		# - = - = - = - = - = - = - = - = - = - = -

		logger.debug(pprint.pformat(d))
		if "multiple_formats" in variables.queue_list[d["info_dict"]["original_url"]]:
			indexx = d["info_dict"]["original_url"] + ":" + d["info_dict"]["format_id"]
		else:
			indexx = d["info_dict"]["original_url"]

		# - = - resolution detector - = - = - = - = - = - = - = - = - = -
		if variables.queue_list[indexx]["resolution"].find("???") > -1 and (variables.queue_list[indexx].get("resolution_detection_tried_on_byte", 0) + 4000000) < int(d.get("downloaded_bytes", 0)) and variables.queue_list[indexx].get("resolution_detection_tries", 0) < 5:
			# int(d["downloaded_bytes"]) > 4000000 # if the file size is too smol, it does not have the needed metadata and ffprobe gives an error
			logger.debug("DOWNBYTES: %s", str(d["downloaded_bytes"]))
			temp1 = get_resolution_ffprobe(d["tmpfilename"])
			temp2 = str(variables.queue_list[indexx].get("resolution_detection_tries", 0)+1)

			if temp1 is not None:
				variables.queue_list[indexx]["resolution"] = temp1
				journal.info(f"[YTCON] Detected resolution: {temp1} (on try {temp2})" )
			else:
				journal.warning(f'[YTCON] Resolution detection failed: ffprobe gave an error (try {temp2})')
			variables.queue_list[indexx]["resolution_detection_tried_on_byte"] = int(d["downloaded_bytes"])
			variables.queue_list[indexx]["resolution_detection_tries"] = variables.queue_list[indexx].get("resolution_detection_tries", 0) + 1
		# - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = -

		variables.queue_list[indexx]["file"] = d["info_dict"]['_filename']
		if variables.queue_list[indexx]["status"] == "exists" and d["status"] == "finished":
			return None # i'll guess it's made to avoid bugs that overwriting some data
		variables.queue_list[indexx]["status"] = d["status"]

		if int(d["_percent_str"].strip().split(".")[0]) > 100:
			journal.warning("[YTCON] yt-dlp returned percent more than 100%: \"" + d["_percent_str"].strip() + "\". Values remain unchanged...")
		else:
			variables.queue_list[indexx]["status_short_display"] = d["_percent_str"].strip()
			variables.queue_list[indexx]["percent"] = d["_percent_str"].strip()
		variables.queue_list[indexx]["speed"] = d["_speed_str"].strip()

		try:
			if variables.queue_list[indexx]["eta"].count(":") > 1:
				variables.queue_list[indexx]["eta"] = d["_eta_str"].strip()
			else:
				variables.queue_list[indexx]["eta"] = "ETA " + d["_eta_str"].strip()
		except KeyError:
			if d["status"] == "finished":
				variables.queue_list[indexx]["eta"] = "ETA 00:00"

		try:
			if d["_total_bytes_estimate_str"].strip() == "N/A":
				variables.queue_list[indexx]["size"] = d["_total_bytes_str"].strip()
			else:
				variables.queue_list[indexx]["size"] = d["_total_bytes_estimate_str"].strip()
		except KeyError:
			pass

		try:
			variables.queue_list[indexx]["downloaded"] = d["_downloaded_bytes_str"].strip()
		except:
			pass

		#logger.debug(pprint.pformat(variables.queue_list))
	except:
		exit_with_exception(traceback.format_exc())
	return None



"""
 * Main init part of app

 * - = - = - =
 * Copyright (C) 2023-2024 Nikita Beloglazov <nnikita.beloglazov@gmail.com>
 *
 * This file is part of github.com/NikitaBeloglazov/ytcon.
 *
 * NikitaBeloglazov/ytcon is free software; you can redistribute it and/or
 * modify it under the terms of the Mozilla Public License 2.0
 * published by the Mozilla Foundation.
 *
 * NikitaBeloglazov/ytcon is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * You should have received a copy of the Mozilla Public License 2.0
 * along with NikitaBeloglazov/ytcon
 * If not, see https://mozilla.org/en-US/MPL/2.0.
"""

import os
import sys

import urwid

#import notify2

debug_that_will_be_saved_later = []
logs_that_will_be_printed_later = []

# - = CHECK SYSTEM PATHS - = - = - = - = - = - = - = - =
# Check if it is android termux emulator and change dir to user-reachable internal storage
this_is_android_device = False
if os.getcwd().find("com.termux") != -1:
	print("[YTCON] Termux not user-reachable directory detected. Changing to /storage/emulated/0...")
	logs_that_will_be_printed_later.append("[YTCON] Termux not user-reachable directory detected.")
	logs_that_will_be_printed_later.append("[YTCON] Changing to /storage/emulated/0...")
	os.chdir("/storage/emulated/0")
	this_is_android_device = True
# - - - - - - - - - - - - -
# Current folder permissions check
try:
	with open("write_test", "wb") as filee:
		pass
	os.remove("write_test")
except:
	print(os.getcwd())
	print("[!!] Current folder is unwritable!")
	if this_is_android_device:
		print("Maybe Termux doesn't have storage permissions?")
	sys.exit(1)
# - - - - - - - - - - - - -
# /tmp folder check (For android and windows compability)
try:
	with open("/tmp/write_test", "wb") as filee:
		pass
	os.remove("/tmp/write_test")
	log_folder = "/tmp/"
except:
	logs_that_will_be_printed_later.append("[YTCON] /tmp folder is unavalible (windows, android?). Set current dir for logs")
	log_folder = ""
# - = - = - = - = - = - = - = - = - = - = - = - = - = - =

from log import init_logger, journal, logger
init_logger(log_folder)

#from control.control import ControlClass
from control.variables import variables
#from control.exit import exit_with_exception, traceback

from render.colors import colors
#from render.progressbar_defs import progressbar_defs
from render.render import render
from render.loop import loop_container
RenderClass = render

from widgets.main_widgets import widgets

from settings.settings_processor import settings

#from settings_menu.variables import settings_menu_variables
from settings_menu.render import sett #, settings_sections

#from app_update import app_updates

#from misc.ffmpeg import get_resolution_ffprobe
from misc.clipboard import clipboard_init#, clipboard_checker

#from downloader.main import downloader
from downloader.hook import hook

# - = - = -

variables.ydl_opts = {
	'logger': journal,
	'progress_hooks': [hook],
	'color': 'no_color',
	#'outtmpl': '%(title)s [%(id)s].%(ext)s', # REALIZED IN own file handler
	'socket_timeout': 15,
	'retries': 20,
	'fragment_retries': 40,
	'retry_sleep': 'http,fragment:exp',
	#'download_archive': 'downloaded_videos.txt', # !!! DANGEROUS OPTION !!! # TODO?
	}

# TODO LIST:
# 'ratelimit': 207200
#
# Angry mode
# 'retries': float("inf"),
# 'fragment_retries': 999\
#
# Desktop notifications

loop_container.loop = urwid.MainLoop(widgets.main_widget, palette=colors.custom_palette)

RenderClass.width, RenderClass.height = loop_container.loop.screen.get_cols_rows()

# - = - = - = - = - = - = -
# Some debug info writer
logger.debug("width: %s", RenderClass.width)
logger.debug("height: %s", RenderClass.height)

# Output collected to-later-print logs
for i in logs_that_will_be_printed_later:
	journal.info(i)
for i in debug_that_will_be_saved_later:
	logger.debug(i)
# - = - = - = - = - = - = -

# - = - = - = - Late initialize - = - = - = - =
settings.load()

if settings.get_setting("clipboard_autopaste") is True:
	clipboard_init()
# - = - = - = - = - = - = - = - = - = - = - = -

from loops import render_tasks
from loops import log_printer
from loops import error_printer

from loops import tick_handlers

loop_container.loop.set_alarm_in(0, render_tasks.render_tasks)
loop_container.loop.set_alarm_in(0, log_printer.log_printer)
loop_container.loop.set_alarm_in(0, error_printer.error_printer)

loop_container.loop.set_alarm_in(0, tick_handlers.tick_handler)
loop_container.loop.set_alarm_in(1, tick_handlers.tick_handler_big_delay)
loop_container.loop.set_alarm_in(1, sett.tick_handler_settings)

# for testing purposes?
# threading.Thread(target=downloader, args=("https://www.youtube.com/watch?v=Kek5Inz-wjQ",), daemon=True).start()

loop_container.loop.run()



"""
	The main component of ytcon, this class sets the basic parameters for the video,
	composes the title and starts downloading.

	For each link one thread (exception: playlists)
"""

import os
import re
import pprint
import threading
import yt_dlp

from log import journal, logger

from control.variables import variables
from control.exit import exit_with_exception, traceback

from settings.settings_processor import settings

from downloader.map_variables import map_variables

from render.progressbar_defs import progressbar_defs

def downloader(url): # pylint: disable=too-many-return-statements
	"""
	The main component of ytcon, this class sets the basic parameters for the video,
	composes the title and starts downloading.

	For each link one thread (exception: playlists)
	"""
	try:
		if url in variables.queue_list:
			if variables.queue_list[url]["status"] not in ("exists", "finished"):
				journal.error(f"[YTCON] Video link \"{progressbar_defs.name_shortener(url, 40)}\" is already downloading!")
				return None

		with yt_dlp.YoutubeDL(variables.ydl_opts) as ydl:
			# needed for some sites. you may need to replace it with the correct one
			if settings.get_setting("special_mode") is True:
				ydl.params["http_headers"]["User-Agent"] = "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36"
				# "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36"
			# - = - = - = Get downloading formats (yt) and generate filename (global) = -
			infolist = ydl.extract_info(url, download=False)

			# - = - = - log spam filter - = - = - = - =
			if infolist is None: # yt-dlp returns videos with errors as None :|| # TODO?
				journal.warning("ydl.extract_info RETURNED NONE", show=False)
				return None
			if "automatic_captions" in infolist:
				del infolist["automatic_captions"]
			if "formats" in infolist:
				del infolist["formats"]
			if "thumbnails" in infolist:
				del infolist["thumbnails"]
			if "heatmap" in infolist:
				del infolist["heatmap"]
			logger.debug(pprint.pformat(infolist))
			# - = - = - = - = - = - = - = - = - = - = -

			# - Playlists support - = - = - = - = - = - = - = - = - = - = -
			if "entries" in infolist:
				for i in infolist["entries"]:
					if i is None: # yt-dlp returns videos with errors as None :||
						continue
					threading.Thread(target=downloader, args=(i["webpage_url"],), daemon=True).start()
				return None
			# - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = -

			# - Name fiter + assemble - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = -
			temp1 = re.sub(r"[^A-Za-z0-9А-Яа-я \-_.,]", "", infolist["title"].replace("&", "and")) # get title, remove all characters except allowed # >"|" -> "｜" yt-dlp, wtf?
			temp1 = " ".join(temp1.removesuffix(" 1").split()) # remove space duplicates and remove 1 in the end because for some reason yt-dlp adds it on its own
			id_in_filename = infolist["id"].removesuffix("-1")
			filename = f'{temp1} [{id_in_filename}].{infolist["ext"]}'

			# Name too long handler (https://github.com/ytdl-org/youtube-dl/issues/29912 and more more more issues)
			if len(filename.encode('utf-8')) > 190:
				# ^^^^^^^^^^^^^^^^^^^^^^^ counting bytes in filename
				logger.debug("ERROR: FILENAME MORE THAN 190 BYTES. SHORTING...")
				while len(filename.encode('utf-8')) > 190:
					temp1 = " ".join(temp1.split()[:-1]) # remove 1 last word
					filename = f'{temp1} [{id_in_filename}].{infolist["ext"]}'
					logger.debug(filename)
					logger.debug(len(filename.encode('utf-8')))
			# - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = -

			# Check if file exists
			exists = os.path.exists(filename)
			if exists:
				journal.warning(f'[YTCON] FILE "{filename}" EXISTS')

			# - = - = - = Set parameters = -
			multiple_formats = False
			if infolist["extractor"] == "youtube" and "requested_formats" in infolist:
				multiple_formats = True

			temp1_index = map_variables.main(multiple_formats, infolist, filename)

			if exists:
				variables.queue_list[infolist["original_url"]]["status"] = "exists"
				if multiple_formats:
					for i in infolist["requested_formats"]:
						temp1_index = infolist["original_url"] + ":" + i["format_id"]
						variables.queue_list[temp1_index]["status"] = "exists"
						variables.queue_list[temp1_index]["downloaded"] = variables.queue_list[temp1_index]["size"]
						variables.queue_list[temp1_index]["status_short_display"] = "Exist"
						variables.queue_list[temp1_index]["percent"] = "100.0%"
				else:
					variables.queue_list[temp1_index]["downloaded"] = variables.queue_list[temp1_index]["size"]
					variables.queue_list[temp1_index]["status_short_display"] = "Exist"
					variables.queue_list[temp1_index]["percent"] = "100.0%"
			# - = - = - = - = - = - = - = - = - = - = - = - =
			logger.debug(pprint.pformat(variables.queue_list))

			with yt_dlp.YoutubeDL(variables.ydl_opts | {"outtmpl": filename}) as ydl2:
				logger.debug(ydl2.download(url))
				if variables.last_error.find("[Errno 36] File name too long") > -1:
					raise yt_dlp.utils.DownloadError(variables.last_error)
			# - = Mark meta as finished = -
			if "meta_index" in variables.queue_list[infolist["original_url"]]:
				variables.queue_list[infolist["original_url"]]["status"] = "finished"

	except yt_dlp.utils.DownloadError as e:
		journal.error(str(e), show=False)
		map_variables.mark_as_error(url)
		return None
	except:
		exit_with_exception(traceback.format_exc())
		return None

	# - = - = - = [Post-processing] = - = - = - #
	try:
		if variables.queue_list[temp1_index]["status"] == "exists":
			return None # skip post-process if file already exists

		# = - = -
		if variables.queue_list[temp1_index]["status"] != "finished":
			# IF DOWNLOAD THREAD EXITS WITHOUT ERROR (this usually occurs due to the "ignoreerrors" flag)
			journal.debug("DOWNLOAD THREAD EXITED WITHOUT ERROR")
			map_variables.mark_as_error(url)
			return None
		# = - = -

		# Removes Last-modified header. Repeats --no-mtime functionality which is not present in yt-dlp embeded version
		os.utime(variables.queue_list[temp1_index]["file"])

		# Remove file after downloading
		if variables.delete_after_download is True:
			journal.warning(f"[YTCON] REMOVING {variables.queue_list[temp1_index]['file']}...")
			os.remove(variables.queue_list[temp1_index]["file"])
	except:
		exit_with_exception(traceback.format_exc())

	return None



"""
This module stores information about the download queue
and some information that must be passed through several functions.
"""

class VariablesStorage:
	"""
	This class stores information about the download queue
	and some information that must be passed through several functions.
	"""
	def __init__(self):
		self.queue_list = {}
		self.ydl_opts = {}

		self.temp = {}
		self.temp["autopaste_button_color"] = "" # some kind of cache, see tick_handler

		self.last_error = ""
		self.error_countdown = 0
		self.prev_last_error = ""
		self.prev_error_countdown = 0

		self.delete_after_download = False

		# TODO: SEE log.py, line 93
		self.log = ["", "", "", "", "", "Logs will appear there.."]

		self.exit = False
		self.auto_update_safe_gui_stop = False
		self.exception = ""
		self.clipboard_checker_state_launched = False

		# See InputHandlerClass.InputBox
		self.alt_plus_arrow_multiline_message_sended = False

variables = VariablesStorage()



""" DEPRECATED: WILL BE DELETED AND REWORKED IN THE FUTURE """
# TODO: delete auto-update, rework detection func

import os
import sys
import time
import pprint
import importlib
import traceback
import threading
import subprocess

import urwid
import requests

from log import journal, logger
from control.variables import variables
from render.colors import colors

class UpdateAndVersionsClass:
	""" The class stores everything related to determining the version number and auto-updates """
	def __init__(self):
		self.version = "?.?.?"
		self.install_source = "???"
		self.pypi_version = "?.?.?"

		self.auto_update_avalible = False

		self.get_pypi_version_new_thread()
		self.version, self.install_source = self.check_version()

		# Widget from settings that shows versions
		self.settings_version_text = urwid.Text((colors.yellow, f"Your YTCON version: {self.version} / Newest YTCON version: *Working..*"))

		# Variables that cannot have initial values but need to be declared
		self.update_thread = None

	def check_version(self):
		""" Tries to determine the installation method. Sometimes using some external tools """
		version = "?.?.?"
		install_source = "???"
		try:
			# Trying to use relative paths to look at __version__
			from .__version__ import __version__ # pylint: disable=import-outside-toplevel
			version = __version__

			# Understand by which tool is it installed
			if os.path.abspath(__file__).find("pipx") > 0:
				install_source = "pipx"
			else:
				# TODO ADD VENV DETECTION
				install_source = "pip"
		except:
			pass

		if version in ("!!{PLACEHOLDER}!!", "?.?.?"):
			try:
				# Find the directory in which the ytcon startup file is located
				ytcon_files_path = os.path.abspath(__file__).replace("app_update.py", "")

				# Try to load __version__.py from the directory where ytcon is located
				spec = importlib.util.spec_from_file_location("version", ytcon_files_path + "__version__.py")
				module = importlib.util.module_from_spec(spec)
				spec.loader.exec_module(module)
				version = module.__version__
			except:
				pass

		if version in ("!!{PLACEHOLDER}!!", "?.?.?"):
			try:
				# Use GIT to check tags, makes sense if git clone was used to install
				tag = subprocess.check_output(f'git -C "{ytcon_files_path}" describe --tags', shell=True, encoding="UTF-8")

				# Formating it a little
				# v0.0.11-3-g0ada3b4 -->> 0.0.11
				tag = tag.replace("\n", "").replace("v", "")
				if tag.find("-") > 1:
					tag = tag[0:tag.find("-")]

				version = tag
				install_source = "git"
			except:
				pass

		if version == "!!{PLACEHOLDER}!!":
			version = "?.?.?"
		return version, install_source

	def get_pypi_version(self):
		""" Get newest version number via PyPI public API """
		try:
			temp1 = requests.get("https://pypi.org/pypi/ytcon/json", timeout=20).json()
			logger.debug(pprint.pformat(temp1))
			self.pypi_version = temp1["info"]["version"]
			#self.pypi_version = "0.0.14"
		except:
			logger.debug(traceback.format_exc())

	def get_pypi_version_new_thread(self):
		""" Just starts a new thread self.get_pypi_version. Made for not to slow down GUI or utility launch """
		threading.Thread(target=self.get_pypi_version, daemon=True).start()

	def update_settings_version_text(self):
		""" Class that generates text for the widget settings_version_text """
		textt = f"Your YTCON version: {self.version} (from {self.install_source}) / Actual YTCON version: {self.pypi_version}"

		if self.version == "?.?.?" or self.pypi_version == "?.?.?":
			self.settings_version_text.set_text((colors.yellow, textt))
		elif self.version == self.pypi_version:
			self.settings_version_text.set_text((colors.green, textt))
		elif self.version != self.pypi_version:
			if self.install_source == "pipx":
				#textt = textt + "\n\nUpdate using pipx:\n - pipx upgrade ytcon\n"
				self.auto_update_avalible = True
			if self.install_source == "pip":
				textt = textt + "\n\nUpdate using pip:\n - pip3 install -U ytcon\n"
			if self.install_source == "git":
				#textt = textt + "\n\nUpdate using git:\n - `git pull` in folder with ytcon\n"
				self.auto_update_avalible = True

			if self.auto_update_avalible is True:
				textt = textt + "\n[!!] Auto update is avalible! Write \"update\" in input field to easy update right now!\n"
			self.settings_version_text.set_text((colors.light_red, textt))

	def get_update_command(self):
		""" A function that stores commands for updating """
		if self.install_source == "pipx":
			return "pipx upgrade ytcon"
		if self.install_source == "git":
			ytcon_files_path = os.path.abspath(__file__).replace("app_update.py", "")
			return f"git -C {ytcon_files_path} pull"
		if self.install_source == "pip":
			journal.error("[YTCON] Auto update is not avalible for pip installation type")
			return None
		if self.install_source == "???":
			journal.error("[YTCON] Auto update is not avalible - Unable to determine your installation type")
			return None
		#else:
		journal.error("[YTCON] Auto update is not avalible - No update instructions found for your installation type")
		return None

	def update_run_and_restart(self):
		""" Starts auto-update """
		update_command = self.get_update_command()
		if update_command is None:
			return None
		variables.auto_update_safe_gui_stop = True
		time.sleep(1)
		print("\n- = - =\nThe following command will run in 10 seconds:")
		print(" - " + update_command)
		print("\nCtrl+C to cancel update.")
		time.sleep(10)
		print("- = - =\n>> " + update_command + "\n")
		status_code = os.system(update_command)
		if status_code == 0:
			print("\n- = - =\nUpdate was completed successfully! YTCON will restart itself..")
			restart_command = sys.executable + " " + " ".join(sys.argv)
			print(" - " + restart_command)
			print("\nCtrl+C to cancel restart.")
			time.sleep(5)
			print("- = - =\n>> " + restart_command + "\n")
			os.system(restart_command)
		else:
			print("\n- = - =\nIt looks like the update failed. See the output above for details.\nYTCON has not been updated.")
		return None

app_updates = UpdateAndVersionsClass()



""" Module that responsible for contolling ffmpeg/ffprobe checks and calls. """
import sys
import subprocess
import pprint
import ffmpeg # | !!! pip install "ffmpeg-python", NOT "ffmpeg" !!! | # https://kkroening.github.io/ffmpeg-python/ # python310-ffmpeg-python

from log import logger

# - = - Check ffmpeg installed in system - = - = - = -
try:
	# Try to launch it
	subprocess.run("ffmpeg -version", shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	subprocess.run("ffprobe -version", shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
except subprocess.CalledProcessError as e:
	# If there was a command execution error, ffmpeg is not installed
	print("\n[!!] FFMPEG and/or FFPROBE is not installed/broken in your system. Install it using system package manager:\n - sudo apt install ffmpeg\n - sudo dnf install ffmpeg\n - sudo zypper install ffmpeg\n - sudo pacman -S ffmpeg.\n\nYTCON execution cannot be continued.\n")
	sys.exit(1)
# - = - = - = - = - = - = - = - = - = - = - = - = - = -

def get_resolution_ffprobe(file):
	""" Uses ffprobe to get video (even not fully downloaded) resolution """
	try:
		probe = ffmpeg.probe(file)
	except ffmpeg.Error as e:
		logger.debug("ffprobe error:")
		logger.debug(e.stderr)
		return None
	logger.debug("ffprobe response:")
	logger.debug(pprint.pformat(probe))
	for i in probe["streams"]:
		if "width" in i and "height" in i:
			return str(i["width"]) + "x" + str(i["height"])
	return None



"""
	Draws errors in widgets.error_widget in red
	After some time (specified in the timer) removes error messages
"""

from log import journal

from control.variables import variables
from control.exit import exit_with_exception, traceback

from render.colors import colors
from render.progressbar_defs import progressbar_defs

from widgets.main_widgets import widgets

# - = - = - = - = - = - = - = - = - =
class ErrorPrinterVariables:
	""" Placeholder """
	def __init__(self):
		""" Placeholder """
		self.animation = 3

error_printer_variables = ErrorPrinterVariables()
# - = - = - = - = - = - = - = - = - =

def error_printer(loop, _):
	""" Draws errors in widgets.error_widget in red, after some time (specified in the timer) removes error messages """
	try:
		# - = skip, do not re-render if there is no errors - = - = - = - = -
		# if variables.prev_last_error == variables.last_error and variables.prev_error_countdown == variables.error_countdown:
		#	time.sleep(0.6)
		#	continue
		# - = - = - = - = - = - = - = - = - = - = - = - = - - = - = - = -
		to_render = []
		to_render.append("- - -\n")

		if variables.error_countdown != 0:
			error_text_generator = "[" + progressbar_defs.whitespace_stabilization(str(variables.error_countdown), 2) + "] " + str(variables.last_error)
		else:
			error_text_generator = str(variables.last_error)

		error_text_generator = error_text_generator.replace("; please report this issue on  https://github.com/yt-dlp/yt-dlp/issues?q= , filling out the appropriate issue template. Confirm you are on the latest version using  yt-dlp -U", "")

		if variables.last_error == "":
			to_render.append((colors.cyan, error_text_generator))
		else:
			to_render.append((colors.red, error_text_generator))

		to_render.append("\n")

		# - = - = - = - = - = - unfold animation - = - = - = - = - = -
		if error_printer_variables.animation == 0:
			widgets.error_widget.set_text(to_render)
		elif error_printer_variables.animation == 1:
			widgets.error_widget.set_text(to_render[:-1])
		elif error_printer_variables.animation == 2:
			if to_render[:-2] == ["- - -\n"]:
				widgets.error_widget.set_text("- - -")
			else:
				widgets.error_widget.set_text(to_render[:-2])
		elif error_printer_variables.animation == 3:
			if not to_render[:-3]:
				widgets.error_widget.set_text("")
			else:
				widgets.error_widget.set_text(to_render[:-3])

		if variables.last_error == "":
			if error_printer_variables.animation < 3 and error_printer_variables.animation >= 0:
				error_printer_variables.animation += 1
		else:
			if error_printer_variables.animation <= 3 and error_printer_variables.animation > 0:
				error_printer_variables.animation = error_printer_variables.animation - 1
		# - = - = - = - = - = - = - = - = - = - = - = - = - = - = -

		variables.prev_last_error = variables.last_error
		variables.prev_error_countdown = variables.error_countdown

		if variables.error_countdown != 0: # TODO: variables.error_countdown might be moved to separate container
			variables.error_countdown = variables.error_countdown - 1
			if variables.error_countdown == 0:
				journal.clear_errors()

		#widgets.error_widget.set_text(to_render)
		loop.set_alarm_in(0.3, error_printer)
	except:
		exit_with_exception(str(traceback.format_exc()))



""" A container with some shared variables for settings menu """

class SettingsMenuVariables:
	""" A container with some shared variables for settings menu """
	def __init__(self):
		self.settings_show = False
		self.settings_showed = False

		self.settings_soft_update_scheduled = False

settings_menu_variables = SettingsMenuVariables()



"""
	A log wrapper for yt-dlp and a logging class for the whole application.
	info, warning and error will be added to the logs field. (logprinter())
	For debug messages it is highly recommended to use logger.debug instead of journal
"""

import logging
from control.variables import variables

from render.render import render

logger = logging.getLogger('main_logger')
def init_logger(log_folder):
	"""
	Logger initialization. Made because imports cannot have argument funcionality,
	Because we can't check log folder in log.py, and it must be detected in yt.py
	"""
	# - = logging init - = - = - = - = - = - = - = - = - = - = - = - =
	logger.setLevel(logging.DEBUG)

	# Create handler for the INFO level
	info_file_handler = logging.FileHandler(log_folder+'info.log', mode='w')
	info_file_handler.setLevel(logging.INFO)

	# Create handler for the DEBUG level
	debug_file_handler = logging.FileHandler(log_folder+'debug.log', mode='w')
	debug_file_handler.setLevel(logging.DEBUG)

	# Add formatter
	formatter = logging.Formatter('%(levelname)s: %(message)s')
	info_file_handler.setFormatter(formatter)
	debug_file_handler.setFormatter(formatter)

	# Add handlers to the logger
	logger.addHandler(info_file_handler)
	logger.addHandler(debug_file_handler)

	# Write test logs
	logger.debug('== DEBUG LOG FILE ==')
	logger.info('== INFO LOG FILE ==')
	# - = - = - = - = - = - = - = - = - = - = - = - = - = - = - = - =

class JournalClass:
	"""
	A log wrapper for yt-dlp and a logging class for the whole application.
	info, warning and error will be added to the logs field. (logprinter())
	For debug messages it is highly recommended to use logger.debug instead of journal
	"""
	def debug(self, msg):
		""" !!! For yt-dlp """
		if msg.startswith('[debug] '):
			logger.debug(msg)
		else:
			self.info(msg)

	# show: show in logs field
	def info(self, msg, show=True):
		""" info log level """
		logger.info(msg)
		if show:
			self.add_to_logs_field(msg)
	def warning(self, msg, show=True):
		""" warning log level """
		logger.warning(msg)
		if show:
			self.add_to_logs_field(msg)
	def error(self, msg, show=True):
		""" error log level. The error message will also be added to the errorprinter() """
		if msg == "ERROR: kwallet-query failed with return code 1. Please consult the kwallet-query man page for details":
			return None # Outdated: does not appear in yt-dlp on python3.11

		logger.error(msg)
		variables.last_error = msg
		variables.error_countdown = 99
		if show:
			self.add_to_logs_field(msg)
		return None

	def clear_errors(self, _=None):
		""" Clear the errorprinter() field """
		variables.last_error = ""
		variables.error_countdown = 0

	def add_to_logs_field(self, msg):
		""" there logs will be added to the logprinter() field. """
		msg = str(msg).replace("\n", "")
		if "[download]" in msg and "%" in msg and "at" in msg:
			# awoid logging such as "[download] 100.0% of   52.05MiB at    3.07MiB/s ETA 00:00"
			return None

		del variables.log[0]

		# TODO: loops.log_printer must cut log messages, not logger. current version is cruel shit
		if len(msg) > render.width:
			temp1 = render.width - 3
			variables.log.append(msg[0:temp1]+"...")
		else:
			variables.log.append(msg)
		# logger.debug(variables.log)
		return None

journal = JournalClass()



""" This module contains some functions for controlling some variables in control.variables. """

from log import journal, logger

from control.variables import variables
from control.exit import exit_with_exception, traceback

from widgets.top_pile import widgets_tp

class ControlClass_base:
	""" This module contains some functions for controlling some variables in control.variables. """
	def delete_finished(self):
		""" Removes all completed operations from ControlClass.queue_list with a loop """
		try:
			temp1 = 0
			temp2_new = variables.queue_list.copy()
			for item, item_content in variables.queue_list.copy().items():
				if item_content["status"] == "exists" or item_content["status"] == "finished" or item_content["status"] == "error":
					del temp2_new[item]
					if "meta_index" not in item_content:
						temp1 = temp1 + 1
			variables.queue_list = temp2_new
			logger.debug(variables.queue_list)
			widgets_tp.top_pile.contents = [] # see render.render.remove_all_widgets
			return str(temp1)
		except:
			exit_with_exception(traceback.format_exc())
		return None

	def clear(self, _=None):
		""" Clears errors and finished downloads from memory """
		journal.clear_errors()
		journal.info(f"[YTCON] {self.delete_finished()} item(s) removed from list!")

	def delete_after_download_switch(self, _=None, _1=None):
		""" Special mode switch function for urwid.Button's """
		# TODO: Move to normal settings
		journal.info("")
		if variables.delete_after_download: # true
			variables.delete_after_download = False
			journal.info("[YTCON] Delete after download disabled!")
		elif not variables.delete_after_download: # false
			variables.delete_after_download = True
			journal.error("[YTCON] Delete after download ENABLED! This means that the downloaded files WILL NOT BE SAVED!")

ControlClass = ControlClass_base()



"""
	Graphic part of ytcon - draws a colored video queue from variables.queue_list
	Shows names, extractors, ETAs, generates progress bars, etc.
"""
from control.variables import variables
from control.exit import exit_with_exception, traceback

from render.colors import colors
from render.progressbar_defs import progressbar_defs
from render.render import render

def render_tasks(loop, _):
	"""
	Graphic part of ytcon - draws a colored video queue from variables.queue_list
	Shows names, extractors, ETAs, generates progress bars, etc.
	"""
	try:
		if not variables.queue_list: # if variables.queue_list == {}
			render.edit_or_add_row((colors.cyan, "No tasks"), 0)
		else:
			r = 0
			for _, i in variables.queue_list.items():
				if "meta_index" in i:
					continue # just ignore meta-downloads

				rcm = progressbar_defs
				ws = rcm.whitespace_stabilization

				errorr = i["status"] == "error"

				temp1 = f'{ws(i["status_short_display"], 7)}{rcm.progressbar_generator(i["percent"], errorr)}{ws(i["speed"], 13)}|{ws(rcm.bettersize(i["downloaded"])+"/"+rcm.bettersize(i["size"]), 15)}| {ws(i["eta"], 9)} | {ws(i["site"], 7)} | {ws(i["resolution"], 9)} | '
				fileshortname = rcm.name_shortener(i["name"], render.width - len(temp1))
				temp1 = temp1 + fileshortname

				if i["status"] == "waiting":
					render.edit_or_add_row((colors.cyan, temp1), r)
				elif i["status"] == "error":
					render.edit_or_add_row((colors.red, temp1), r)
				elif i["status"] == "exists":
					render.edit_or_add_row((colors.yellow, temp1), r)
				elif i["status"] == "finished":
					render.edit_or_add_row((colors.green, temp1), r)
				else:
					render.edit_or_add_row(temp1, r)

				r = r+1
		loop.set_alarm_in(0.3, render_tasks)
	except:
		exit_with_exception(traceback.format_exc())



""" Contains some static methods for rendering """
import time

from render.loop import loop_container

class RenderStatic:
	""" Contains some static methods for rendering """
	def __init__(self):
		pass

	def flash_button_text(self, button, color, times=4):
		""" Makes the button to blink in the specified color """
		if button is None:
			return None
		temp1 = button.get_label()
		for _ in range(1, times+1):
			button.set_label((color, temp1))
			loop_container.loop.draw_screen()
			time.sleep(0.1)
			button.set_label(temp1)
			loop_container.loop.draw_screen()
			time.sleep(0.1)
		return None

render_static = RenderStatic()



""" Stops the urwid and prints an error message, then exits the program """

import traceback
# TODO: catch traceback by self, like
"""
>>> def lol(): print(traceback.format_exc()); print("yes")
...
>>> try:
...     0/0
... except:
...     lol()
...
Traceback (most recent call last):
File "<stdin>", line 2, in <module>
ZeroDivisionError: division by zero
yes
"""

from log import journal
from control.variables import variables

# TODO: Consider stop by self, using loop_container
def exit_with_exception(text):
	""" Stops the urwid and prints an error message, then exits the program """
	journal.error(text)
	variables.exit = True
	variables.exception = text + "\nPrinted by exit_with_exception()"



"""
	This module stores some information about rendering, screen,
	some functions for working with widgets
	and some functions that are related to rendering.
"""
import os
import urwid

#from widgets.main_widgets import widgets
from widgets.top_pile import widgets_tp

class RenderClass:
	""" It stores some information about rendering, screen, some functions for working with widgets and some functions that are related to rendering. """
	def __init__(self):
		# We need to have some numbers for initializing,
		# but loop_container.loop.screen.get_cols_rows() is not ready.
		# so i use os.get_terminal_size(0) for some numbers to init.
		self.width, self.height = os.get_terminal_size(0)

	def add_row(self, text):
		""" Add an additional widget to top_pile for drawing a new task """
		widgets_tp.top_pile.contents = widgets_tp.top_pile.contents + [[urwid.Text(text), widgets_tp.top_pile.options()],]

	def edit_or_add_row(self, text, pos):
		""" Edit a widget with a specific serial number, and if there is none, then create a new one """
		if pos > self.calculate_widget_height(widgets_tp.top_pile) - 1:
			self.add_row(text)
		else:
			widgets_tp.top_pile.contents[pos][0].set_text(text)

	#def remove_all_widgets(self):
	#	- = + DEPRECATED + = -
	#	USE widgets_tp.top_pile.contents = [] UNSTEAD
	#
	#	"""
	#	If there are obsolete widgets in top_pile that will not be changed, they are considered garbage,
	#	for this you need to call remove_all_widgets, all widgets, including unnecessary old ones,
	#	will be removed, but will be recreated if needed
	#	"""
	#	widgets_tp.top_pile.contents = []

	def calculate_widget_height(self, widget):
		""" (recursively) Counts how many rows the widget occupies in height """
		if isinstance(widget, urwid.Text):
			# Returns the number of lines of text in the widget
			return len(widget.text.split('\n'))
		if isinstance(widget, urwid.Pile):
			# Recursively sums the heights of widgets inside a urwid.Pile container
			return sum(self.calculate_widget_height(item[0]) for item in widget.contents)
		return 0 # Return 0 for unsupported widget types (?)

render = RenderClass()



""" LoopContainer - container class that shares iniated loop variable (urwid.MainLoop) """

class LoopContainer:
	"""" LoopContainer - container class that shares iniated loop variable (urwid.MainLoop) """
	def __init__(self):
		self.loop = None

loop_container = LoopContainer()



""" Prints the last 6 lines of logs in widgets.log_widget """
from control.variables import variables
from control.exit import exit_with_exception, traceback

from widgets.main_widgets import widgets

def log_printer(loop, _): # TODO: POSSIBLE, NEEDS REWRITE/BUGFIX. SEE log.py, line 93
	""" Prints the last 6 lines of logs in widgets.log_widget """
	try:
		# - = skip, do not re-render if it doesn't change - = - = - =
		# if ControlClass.oldlog == variables.log:
		#	time.sleep(0.5)
		#	continue
		# else:
		#	ControlClass.oldlog = variables.log.copy()
		#
		# controlclass snippet:
		# self.oldlog = ["", "", "", "", "", ""]
		# - = - = - = - = - = - = - = - = - = - = - = - = - - = - = - =

		to_render = variables.log[0] + "\n"
		to_render += variables.log[1] + "\n"
		to_render += variables.log[2] + "\n"
		to_render += variables.log[3] + "\n"
		to_render += variables.log[4] + "\n"
		to_render += variables.log[5]
		widgets.log_widget.set_text(to_render)

		loop.set_alarm_in(0.3, log_printer)
	except:
		exit_with_exception(traceback.format_exc())



"""
	Class for processing user input.
	Contains a modified Urwid.edit widget and functions for processing commands and URLs
"""
import pprint
import traceback
import threading

import urwid

from log import journal, logger
#from render.render import render
from settings_menu.variables import settings_menu_variables
from render.loop import loop_container
from control.variables import variables
from control.control import ControlClass
from control.exit import exit_with_exception, traceback
from settings.settings_processor import settings

from app_update import app_updates

from downloader.main import downloader

class InputHandlerClass:
	"""
		Class for processing user input.
		Contains a modified Urwid.edit widget and functions for processing commands and URLs
	"""
	class InputBox(urwid.Edit):
		"""
			A modified urwid.Edit Widget.
			If the user presses Enter, it collects text and sent text to input_handler,
			and after that is it cleans the input field
		"""
		def is_skipable(self, inp):
			"""
				Determines whether this character can be skipped.
				Used in ALT key handlers to determine whether a character is part of a word
			"""
			if inp.isalpha() or inp.isdigit():# or inp == "%":
				return True
			return False

		def get_cords(self, size):
			"""
				Takes the cursor coordinates from the InputBox to determine which character the cursor is currently on.
				Doesn't work well if the text is muiti-line
			"""
			tmp1 = self.get_cursor_coords(size)[0]
			if len(self.get_cursor_coords(size)) > 1 and self.get_cursor_coords(size)[1] > 0:
				if variables.alt_plus_arrow_multiline_message_sended is False:
					journal.error("[YTCON] Navigation using Alt+Arrow in a multi-line input field does not work as expected, this is a known problem, and we cannot solve it due to the peculiarities of the engine.\n\nNavigation will be inaccurate and skip some characters for no reason.\nWe apologize for the inconvenience caused.")
					variables.alt_plus_arrow_multiline_message_sended = True
				tmp1 = tmp1 * self.get_cursor_coords(size)[1]
				return tmp1
			return tmp1 - 12

		def get_safe_text(self):
			""" Limits the end of the text to avoid infinite recursion """
			return self.get_edit_text() + "??"

		def keypress(self, size, key):
			""" Overrides a regular class. """
			#journal.info(key)

			if key in ('meta left', 'ctrl left'):
				# Alt + Left and Ctrl + Left key handler. Moves left one word
				super().keypress(size, "left")
				temp1 = ""
				# Moves one letter at a time until it finds a special symbol that cannot an part of word
				while self.is_skipable(self.get_safe_text()[self.get_cords(size)-1]):# and self.get_cords(size) > 0:
					endless_loop_detector_first = self.get_cords(size)
					temp1 = temp1 + self.get_safe_text()[self.get_cords(size)-1]
					super().keypress(size, "left")

					# If the coordinates have not changed since the last cursor movement, then the border has been reached
					endless_loop_detector_two = self.get_cords(size)
					if endless_loop_detector_first == endless_loop_detector_two:
						logger.debug("meta left: loop detected")
						break
				return None

			if key in ('meta right', 'ctrl right'):
				# Alt + Right and Ctrl + Right key handler. Moves right one word
				super().keypress(size, "right")
				temp1 = ""
				# Moves one letter at a time until it finds a special symbol that cannot an part of word
				while self.is_skipable(self.get_safe_text()[self.get_cords(size)]):# and self.get_cords(size) > 0:
					endless_loop_detector_first = self.get_cords(size)
					temp1 = temp1 + self.get_safe_text()[self.get_cords(size)]
					super().keypress(size, "right")

					# If the coordinates have not changed since the last cursor movement, then the border has been reached
					endless_loop_detector_two = self.get_cords(size)
					if endless_loop_detector_first == endless_loop_detector_two:
						logger.debug("meta right: loop detected")
						break
				return None

			if key == 'meta backspace':
				# Alt + Backspace key handler. Removes last word in inputbox
				if len(self.get_edit_text()) < 3:
					super().keypress(size, "backspace")
					return None

				temp1 = list(self.get_edit_text())
				temp2 = False
				if self.is_skipable(temp1[-1]):
					temp2 = True
				else:
					if self.is_skipable(temp1[-2]):
						temp2 = True
					del temp1[-1]

				while self.is_skipable(temp1[-1]) and temp2 is True:
					del temp1[-1]
					if len(temp1) == 0:
						break

				self.set_edit_text("".join(temp1))
				return None

			if key == 'enter':
				# If enter pressed, send URL to input_handler and clear inputbox
				InputHandler.input_handler(self.get_edit_text())
				self.set_edit_text("")
				return None

			# If the conditions do not work and the key is not assigned
			return super().keypress(size, key)

	class InputProcessed(Exception):
		""" Dummy exception, when called means that the processing of this request is completed. """

	def input_handler(self, text):
		""" Main input handler logic """
		try:
			original_text = text
			text = text.lower()

			if text == "":
				# Force refreshing screen...
				loop_container.loop.draw_screen()
				raise self.InputProcessed

			journal.info("")
			journal.info("[YTCON] [INPUT] " + original_text)

			# - = Clipboard auto-paste = -
			if text in ("cb", "clipboard", "clip"):
				settings.clipboard_autopaste_switch()
				raise self.InputProcessed
			# - = - = - = - = - = - = - =

			# - = Delete after download = -
			if text in ("dad", "delete after download"):
				ControlClass.delete_after_download_switch()
				raise self.InputProcessed
			# - = - = - = - = - = - = - =

			if text in ("clear", "cls"):
				ControlClass.clear()

			elif text == "logtest":
				logger.debug("[TEST] 1")
				journal.info("[TEST] 2")
				journal.warning("[TEST] 3")
				journal.error("[TEST] 4")
				journal.error("[TEST] 5", show=False)
				journal.info("😘😘😘😘 6") # can break something, emojis have problems calculating sizes

			elif text == "crash":
				try:
					0/0
				except:
					exit_with_exception(traceback.format_exc())

			elif text == "s":
				settings_menu_variables.settings_show = True

			elif text == "flags":
				journal.info(pprint.pformat(variables.ydl_opts))

			elif text == "s ls":
				journal.info(settings.settings)

			elif text == "save":
				settings.save()
				#journal.info(settings.settings)
			elif text == "load":
				settings.load()
				#journal.info(settings.settings)

			elif text == "update":
				#app_updates.update_run_and_restart()
				app_updates.update_thread = threading.Thread(target=app_updates.update_run_and_restart, daemon=True)
				app_updates.update_thread.start()

			elif text == "fake update":
				app_updates.pypi_version = "0.0.99"

			else:
				threading.Thread(target=downloader, args=(original_text,), daemon=True).start()

		except self.InputProcessed:
			pass
		except:
			exit_with_exception(traceback.format_exc())

InputHandler = InputHandlerClass()



""" This module stores some information about colors and some color presets """
import urwid

class ColorsBase:
	""" Stores some information about colors and some color presets. """
	def __init__(self):
		# Init colors
		self.red = urwid.AttrSpec('dark red', 'default')
		self.light_red = urwid.AttrSpec('light red', 'default')
		self.yellow = urwid.AttrSpec('brown', 'default')
		self.light_yellow = urwid.AttrSpec('yellow', 'default')
		self.green = urwid.AttrSpec('dark green', 'default')
		self.cyan = urwid.AttrSpec('dark cyan', 'default')
		# self.light_white = urwid.AttrSpec('bold', 'default') # unused

		self.custom_palette = [
			# ('name_of_style', 'color_text', 'color_background')
			('reversed', 'standout', ''),
			('buttons_footer', 'light green', ''),
			('light_red', 'light red', ''),
			('yellow', 'yellow', ''),
			('light_green', 'light green', ''),
			('green', 'dark green', ''),
			('cyan', 'dark cyan', ''),
			('light_cyan', 'light cyan', ''),
			('bold_default', 'white', ''),

			('green_background', 'black', 'dark green'),
			('cyan_background',  'black', 'dark cyan'),
		]

colors = ColorsBase()
